import logging
from collections.abc import Sequence
from typing import Literal

import numpy as np
from anndata import AnnData
from scvi import REGISTRY_KEYS, settings
from scvi.data import AnnDataManager
from scvi.data.fields import CategoricalObsField, LayerField, NumericalJointObsField
from scvi.model.base import BaseModelClass, UnsupervisedTrainingMixin, VAEMixin
from scvi.utils import setup_anndata_dsp

import drvi
from drvi.nn_modules.feature_interface import FeatureInfoList
from drvi.scvi_tools_based.data.fields import FixedCategoricalJointObsField
from drvi.scvi_tools_based.merlin_data import (
    MerlinData,
    MerlinDataManager,
    MerlinDataSplitter,
    MerlinTransformedDataLoader,
)
from drvi.scvi_tools_based.merlin_data import (
    fields as melin_fields,
)
from drvi.scvi_tools_based.model.base import DRVIArchesMixin, GenerativeMixin
from drvi.scvi_tools_based.module import DRVIModule

logger = logging.getLogger(__name__)


class DRVI(VAEMixin, DRVIArchesMixin, UnsupervisedTrainingMixin, BaseModelClass, GenerativeMixin):
    """
    DRVI model based on scvi-tools skelethon

    Parameters
    ----------
    adata
        AnnData object that has been registered via :meth:`~drvi.model.DRVI.setup_anndata`.
    n_latent
        Dimensionality of the latent space.
    encoder_dims
        Number of nodes in hidden layers of the encoder.
    decoder_dims
        Number of nodes in hidden layers of the decoder.
    prior
        Prior model. defaults to normal.
    prior_init_obs
        When initializing priors from data, these observations are used.
    categorical_covariates
        Categorical Covariates as a list of texts. You can specify emb dimension by appending @dim to each cpvariate.
    **model_kwargs
        Keyword args for :class:`~drvi.model.DRVIModule`

    Examples
    --------
    >>> adata = anndata.read_h5ad(path_to_anndata)
    >>> drvi.DRVI.setup_anndata(adata, categorical_covariate_keys=["batch"])
    >>> vae = drvi.DRVI(adata)
    >>> vae.train()
    >>> adata.obsm["latent"] = vae.get_latent_representation()
    """

    def __init__(
        self,
        adata: AnnData | MerlinData,
        n_latent: int = 32,
        encoder_dims: Sequence[int] = (128, 128),
        decoder_dims: Sequence[int] = (128, 128),
        prior: Literal["normal", "gmm_x", "vamp_x"] = "normal",
        prior_init_obs: np.ndarray | None = None,
        categorical_covariates: list[str] = (),
        **model_kwargs,
    ):
        super().__init__(adata)

        # TODO: Remove later. Currently used to detect autoreload problems sooner.
        if isinstance(adata, AnnData):
            pass
        elif MerlinData is not None and isinstance(adata, MerlinData):
            self._data_splitter_cls = MerlinDataSplitter
        else:
            raise ValueError(
                "Only AnnData and MerlinData are supported. "
                "If you have passed an instalce of MerlinData and still get this error, "
                "make sure merlin is installed as a dependency."
            )

        categorical_covariates_info = FeatureInfoList(categorical_covariates, axis="obs", default_dim=10)
        if REGISTRY_KEYS.CAT_COVS_KEY in self.adata_manager.data_registry:
            cat_cov_stats = self.adata_manager.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY)
            n_cats_per_cov = cat_cov_stats.n_cats_per_key
            assert tuple(categorical_covariates_info.names) == tuple(cat_cov_stats.field_keys)
        else:
            n_cats_per_cov = []
            assert len(categorical_covariates_info) == 0
        n_continuous_cov = self.summary_stats.get("n_extra_continuous_covs", 0)

        prior_init_dataloader = None
        if prior_init_obs is not None:
            assert "_" in prior
            assert len(prior_init_obs) == int(prior.split("_")[-1])
            prior_init_dataloader = self._make_data_loader(
                adata=adata[prior_init_obs], batch_size=len(prior_init_obs), shuffle=False
            )

        self.module = DRVIModule(
            n_input=self.summary_stats["n_vars"],
            n_latent=n_latent,
            encoder_dims=encoder_dims,
            decoder_dims=decoder_dims,
            n_cats_per_cov=n_cats_per_cov,
            n_continuous_cov=n_continuous_cov,
            prior=prior,
            prior_init_dataloader=prior_init_dataloader,
            categorical_covariate_dims=categorical_covariates_info.dims,
            **model_kwargs,
        )

        self._model_summary_string = (
            "DRVI \n"
            + (f"Covariates: {categorical_covariates_info.names}, \n" if len(categorical_covariates_info) > 0 else "")
            + f"Latent size: {self.module.n_latent}, "
            f"splits: {self.module.n_split_latent}, "
            f"pooling of splits: '{self.module.split_aggregation}', \n"
            f"Encoder dims: {encoder_dims}, \n"
            f"Decoder dims: {decoder_dims}, \n"
            f"Gene likelihood: {self.module.gene_likelihood}, \n"
        )
        # necessary line to get params that will be used for saving/loading
        self.init_params_ = self._get_init_params(locals())

        logger.info("The model has been initialized")

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        labels_key: str | None = None,
        layer: str | None = None,
        is_count_data: bool = True,
        categorical_covariate_keys: list[str] | None = None,
        continuous_covariate_keys: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        %(summary)s.

        Parameters
        ----------
        %(param_adata)s
        %(param_labels_key)s
        %(param_layer)s
        %(param_cat_cov_keys)s
        %(param_cont_cov_keys)s

        Returns
        -------
        %(returns)s
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        setup_method_args["drvi_version"] = drvi.__version__

        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=is_count_data),
            CategoricalObsField(REGISTRY_KEYS.LABELS_KEY, labels_key),
            FixedCategoricalJointObsField(REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys),
            NumericalJointObsField(REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    @classmethod
    def setup_merlin_data(
        cls,
        merlin_data: MerlinData,
        labels_key: str | None = None,
        layer: str = "X",
        is_count_data: bool = True,
        categorical_covariate_keys: list[str] | None = None,
        continuous_covariate_keys: list[str] | None = None,
        **kwargs,
    ):
        setup_method_args = cls._get_setup_method_args(**locals())
        setup_method_args["drvi_version"] = drvi.__version__

        fields = [
            melin_fields.MerlinLayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=is_count_data),
            melin_fields.MerlinCategoricalObsField(REGISTRY_KEYS.LABELS_KEY, labels_key),
            melin_fields.MerlinCategoricalJointObsField(REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys),
            melin_fields.MerlinNumericalJointObsField(REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys),
        ]
        merlin_manager = MerlinDataManager(fields, setup_method_args=setup_method_args)
        merlin_manager.register_fields(merlin_data, **kwargs)
        cls.register_manager(merlin_manager)

    def _make_data_loader(
        self,
        adata,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        shuffle: bool = False,
        data_loader_class=None,
        **data_loader_kwargs,
    ):
        """Create a AnnDataLoader object for data iteration.

        Parameters
        ----------
        adata
            AnnData or MerlinData object with equivalent structure to initial AnnData.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        shuffle
            Whether observations are shuffled each iteration though
        data_loader_class
            Class to use for data loader
        data_loader_kwargs
            Kwargs to the class-specific data loader class
        """
        if isinstance(adata, AnnData):
            return super()._make_data_loader(
                adata, indices, batch_size, shuffle, data_loader_class, **data_loader_kwargs
            )
        elif MerlinData is not None and isinstance(adata, MerlinData):
            adata_manager = self.get_anndata_manager(adata)
            if adata_manager is None:
                raise AssertionError(
                    "AnnDataManager not found. Call `self._validate_anndata` prior to calling this function."
                )
            if batch_size is None:
                batch_size = settings.batch_size
            return MerlinTransformedDataLoader(
                self.adata_manager.get_dataset("default"),
                mapping=self.adata_manager.get_fields_schema_mapping(),
                batch_size=batch_size,
                shuffle=shuffle,
                parts_per_chunk=1,
                **data_loader_kwargs,
            )
        else:
            raise ValueError(
                "Only AnnData and MerlinData are supported. "
                "If you have passed an instalce of MerlinData and still get this error, "
                "make sure merlin is installed as a dependency."
            )



import importlib


def get_placeholder(name, error_message, allow_init=False):
    error_message = error_message + f" Cannot use '{name}'."

    class ClassLevelGetAttrMeta(type):
        def __getattr__(cls, name):
            raise ImportError(error_message)

    class LazyNonExistingModulePlaceholder(metaclass=ClassLevelGetAttrMeta):
        def __init__(self):
            if not allow_init:
                raise ImportError(error_message)
            super().__init__()

    return LazyNonExistingModulePlaceholder


if importlib.util.find_spec("merlin"):
    from . import fields
    from ._data import MerlinData
    from ._data_loader import MerlinTransformedDataLoader
    from ._data_manager import MerlinDataManager
    from ._data_splitter import MerlinDataSplitter
else:
    error_msg = "Merlin is not installed. To use merline dataloader please install it."
    fields = get_placeholder("fields", error_msg)
    MerlinData = get_placeholder("MerlinData", error_msg)
    MerlinTransformedDataLoader = get_placeholder("MerlinTransformedDataLoader", error_msg)
    MerlinDataManager = get_placeholder("MerlinDataManager", error_msg)
    MerlinDataSplitter = get_placeholder("MerlinDataSplitter", error_msg)


__all__ = [
    "MerlinData",
    "MerlinTransformedDataLoader",
    "MerlinDataManager",
    "MerlinDataSplitter",
    "fields",
]



from drvi.nn_modules.feature_interface import FeatureInfo, FeatureInfoList


class TestFeatureInfoList:
    def test_fill_with_total_dim(self):
        fil = FeatureInfoList(["x@16!tada", "y", "z!zizi!gulu", "w@8"], total_dim=32)
        assert fil.feature_info_list[0] == FeatureInfo("x", 16, ("tada",))
        assert fil.feature_info_list[1] == FeatureInfo("y", 4, ())
        assert fil.feature_info_list[2] == FeatureInfo("z", 4, ("zizi", "gulu"))
        assert fil.feature_info_list[3] == FeatureInfo("w", 8, ())

        assert tuple(fil.names) == ("x", "y", "z", "w")
        assert tuple(fil.dims) == (16, 4, 4, 8)
        assert tuple(fil.keywords_list) == (("tada",), (), ("zizi", "gulu"), ())



from torch import nn

from drvi.nn_modules.layer.linear_layer import StackedLinearLayer
from drvi.nn_modules.layer.structures import SimpleResidual


class LayerFactory:
    def __init__(self, intermediate_arch="SAME", residual_preferred=False):
        assert intermediate_arch in ["SAME", "FC"]

        self.intermediate_arch = intermediate_arch
        self.residual_preferred = residual_preferred

    def _get_normal_layer(self, d_in, d_out, bias=True, **kwargs):
        raise NotImplementedError()

    def _get_stacked_layer(self, d_channel, d_in, d_out, bias=True, **kwargs):
        raise NotImplementedError()

    def get_normal_layer(self, d_in, d_out, bias=True, intermediate_layer=None, **kwargs):
        if intermediate_layer is None:
            intermediate_layer = True
        if intermediate_layer and self.intermediate_arch == "FC":
            layer = nn.Linear(d_in, d_out, bias)
        elif (not intermediate_layer) or self.intermediate_arch == "SAME":
            layer = self._get_normal_layer(d_in, d_out, bias=True, **kwargs)
        else:
            raise NotImplementedError()

        if self.residual_preferred and (d_in == d_out):
            layer = SimpleResidual(layer)
        return layer

    def get_stacked_layer(self, d_channel, d_in, d_out, bias=True, intermediate_layer=None, **kwargs):
        if intermediate_layer is None:
            intermediate_layer = True
        if intermediate_layer and self.intermediate_arch == "FC":
            layer = StackedLinearLayer(d_channel, d_in, d_out, bias)
        elif (not intermediate_layer) or self.intermediate_arch == "SAME":
            layer = self._get_stacked_layer(d_channel, d_in, d_out, bias=True, **kwargs)
        else:
            raise NotImplementedError()

        if self.residual_preferred and (d_in == d_out):
            layer = SimpleResidual(layer)
        return layer


class FCLayerFactory(LayerFactory):
    def __init__(self, intermediate_arch="SAME", residual_preferred=False):
        super().__init__(intermediate_arch=intermediate_arch, residual_preferred=residual_preferred)

    def _get_normal_layer(self, d_in, d_out, bias=True, **kwargs):
        return nn.Linear(d_in, d_out, bias=bias)

    def _get_stacked_layer(self, d_channel, d_in, d_out, bias=True, **kwargs):
        return StackedLinearLayer(d_channel, d_in, d_out, bias=bias)

    def __str__(self):
        return f"FCLayerFactory(residual_preferred={self.residual_preferred})"



from anndata import AnnData
from pandas.api.types import CategoricalDtype
from scvi.data._utils import _make_column_categorical
from scvi.data.fields import CategoricalJointObsField


class FixedCategoricalJointObsField(CategoricalJointObsField):
    """An AnnDataField for a collection of categorical .obs fields in the AnnData data structure."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _make_array_categorical(self, adata: AnnData, category_dict: dict[str, list[str]] | None = None) -> dict:
        """Make the .obsm categorical."""
        if self.attr_keys != getattr(adata, self.attr_name)[self.attr_key].columns.tolist():
            raise ValueError(
                f"Original .{self.source_attr_name} keys do not match the columns in the ",
                f"generated .{self.attr_name} field.",
            )

        categories = {}
        df = getattr(adata, self.attr_name)[self.attr_key]
        for key in self.attr_keys:
            categorical_dtype = (
                # TODO: make a PR for this fix
                # Only the following line (ordered=True) is changed in the whole function
                CategoricalDtype(categories=category_dict[key], ordered=True) if category_dict is not None else None
            )
            mapping = _make_column_categorical(df, key, key, categorical_dtype=categorical_dtype)
            categories[key] = mapping

        store_cats = categories if category_dict is None else category_dict

        mappings_dict = self._default_mappings_dict()
        mappings_dict[self.MAPPINGS_KEY] = store_cats
        mappings_dict[self.FIELD_KEYS_KEY] = self.attr_keys
        for k in self.attr_keys:
            mappings_dict[self.N_CATS_PER_KEY].append(len(store_cats[k]))
        return mappings_dict



import scanpy as sc


def hvg_batch(
    adata,
    batch_key=None,
    target_genes=2000,
    flavor="cell_ranger",
    n_bins=20,
    adataOut=False,
):
    """
    Batch-aware highly variable gene selection

    Method to select HVGs based on mean dispersions of genes that are highly
    variable genes in all batches. Using a the top target_genes per batch by
    average normalize dispersion. If target genes still hasn't been reached,
    then HVGs in all but one batches are used to fill up. This is continued
    until HVGs in a single batch are considered.

    :param adata: ``anndata`` object
    :param batch: ``adata.obs`` column
    :param target_genes: maximum number of genes (intersection reduces the number of genes)
    :param flavor: parameter for ``scanpy.pp.highly_variable_genes``
    :param n_bins: parameter for ``scanpy.pp.highly_variable_genes``
    :param adataOut: whether to return an ``anndata`` object or a list of highly variable genes
    """
    print("This function is obtained from scIB metrics. Please don't forget to cite them if you use it.")

    adata_hvg = adata if adataOut else adata.copy()

    n_batches = len(adata_hvg.obs[batch_key].cat.categories)

    # Calculate double target genes per dataset
    sc.pp.highly_variable_genes(
        adata_hvg,
        flavor=flavor,
        n_top_genes=target_genes,
        n_bins=n_bins,
        batch_key=batch_key,
    )

    nbatch1_dispersions = adata_hvg.var["dispersions_norm"][
        adata_hvg.var.highly_variable_nbatches > len(adata_hvg.obs[batch_key].cat.categories) - 1
    ]

    nbatch1_dispersions.sort_values(ascending=False, inplace=True)

    if len(nbatch1_dispersions) > target_genes:
        hvg = nbatch1_dispersions.index[:target_genes]

    else:
        enough = False
        print(f"Using {len(nbatch1_dispersions)} HVGs from full intersect set")
        hvg = nbatch1_dispersions.index[:]
        not_n_batches = 1

        while not enough:
            target_genes_diff = target_genes - len(hvg)

            tmp_dispersions = adata_hvg.var["dispersions_norm"][
                adata_hvg.var.highly_variable_nbatches == (n_batches - not_n_batches)
            ]

            if len(tmp_dispersions) < target_genes_diff:
                print(f"Using {len(tmp_dispersions)} HVGs from n_batch-{not_n_batches} set")
                hvg = hvg.append(tmp_dispersions.index)
                not_n_batches += 1

            else:
                print(f"Using {target_genes_diff} HVGs from n_batch-{not_n_batches} set")
                tmp_dispersions.sort_values(ascending=False, inplace=True)
                hvg = hvg.append(tmp_dispersions.index[:target_genes_diff])
                enough = True

    print(f"Using {len(hvg)} HVGs")

    if not adataOut:
        del adata_hvg
        return hvg.tolist()
    else:
        return adata_hvg[:, hvg].copy()



from typing import Literal

import torch
from scvi.distributions import NegativeBinomial
from torch.distributions import Distribution, Normal, Poisson
from torch.nn import functional as F


class NoiseModel:
    def __init__(self):
        pass

    @property
    def parameters(self):
        raise NotImplementedError()

    @property
    def main_param(self):
        return "mean"

    def initial_transformation(self, x, x_mask=1.0):
        x = x
        aux_info = {}
        return x, aux_info

    def dist(self, aux_info, parameters, lib_y):
        # lib_y is so strange here. I should think about it.
        raise NotImplementedError()


def calculate_library_size(x, x_mask=1.0):
    return (x_mask * x).sum(dim=-1) if x_mask is not None else x.sum(dim=-1)


def library_size_normalization(
    x, lib_size, library_normalization: Literal["none", "x_lib", "x_loglib", "div_lib_x_loglib", "x_loglib_all"]
):
    # TODO: remove any library_normalization but 'none', 'x_lib'
    if library_normalization in ["none", "x_loglib"]:
        x = x
    elif library_normalization in ["x_lib", "div_lib_x_loglib"]:
        x = x / lib_size.unsqueeze(-1) * 1e4
    elif library_normalization in ["x_loglib_all"]:
        x = x / torch.log(lib_size.unsqueeze(-1)) * 1e1
    else:
        raise NotImplementedError()
    return x


def library_size_correction(
    x,
    lib_size,
    library_normalization: Literal["none", "x_lib", "x_loglib", "div_lib_x_loglib", "x_loglib_all"],
    log_space=False,
):
    # TODO: remove any library_normalization but 'none', 'x_lib'
    if library_normalization in ["none"]:
        x = x
    elif library_normalization in ["x_lib"]:
        if not log_space:
            x = x * lib_size.unsqueeze(-1).clip(1) / 1e4
        else:
            x = x + torch.log(lib_size.unsqueeze(-1) / 1e4).clip(0)
    elif library_normalization in ["x_loglib", "div_lib_x_loglib", "x_loglib_all"]:
        if not log_space:
            x = x * torch.log(lib_size.unsqueeze(-1)) / 1e4
        else:
            x = x + torch.log(torch.log(lib_size.unsqueeze(-1)).clip(0) / 1e4).clip(0)
    else:
        raise NotImplementedError()
    return x


class NormalNoiseModel(NoiseModel):
    def __init__(self, model_var="fixed", eps=1e-8):
        super().__init__()
        self.model_var = model_var
        self.eps = eps

    @property
    def parameters(self):
        if self.model_var == "fixed":
            var_desc = "fixed=1e-2"
        elif self.model_var.startswith("fixed="):
            var_desc = self.model_var
        elif self.model_var == "dynamic":
            var_desc = "no_transformation"
        elif self.model_var == "feature":
            var_desc = "per_feature"
        else:
            raise NotImplementedError()
        return {
            "mean": "no_transformation",
            "var": var_desc,
        }

    def initial_transformation(self, x, x_mask=1.0):
        x = x
        aux_info = {}
        return x, aux_info

    def dist(self, aux_info, parameters, lib_y):
        var = parameters["var"]
        if self.model_var:
            var = torch.nan_to_num(torch.exp(var), posinf=100, neginf=0) + self.eps
        return Normal(parameters["mean"], torch.abs(var).sqrt())


class PoissonNoiseModel(NoiseModel):
    def __init__(self, mean_transformation="exp", library_normalization="x_lib"):
        super().__init__()
        self.mean_transformation = mean_transformation
        self.library_normalization = library_normalization

    @property
    def parameters(self):
        return {
            "mean": "no_transformation",
        }

    def initial_transformation(self, x, x_mask=1.0):
        x = x
        aux_info = {}
        aux_info["x_library_size"] = calculate_library_size(x, x_mask)
        x = library_size_normalization(x, aux_info["x_library_size"], self.library_normalization)
        x = torch.log1p(x)
        return x, aux_info

    def dist(self, aux_info, parameters, lib_y):
        mean = parameters["mean"]
        library_size = lib_y

        if self.mean_transformation == "exp":
            trans_mean = torch.exp(mean)
            trans_mean = library_size_correction(trans_mean, library_size, self.library_normalization, log_space=False)
        elif self.mean_transformation == "softmax":
            trans_mean = torch.softmax(mean, dim=-1)
            trans_mean = library_size.unsqueeze(-1) * trans_mean
        else:
            raise NotImplementedError()
        return Poisson(trans_mean)


class NegativeBinomialNoiseModel(NoiseModel):
    def __init__(self, dispersion="feature", mean_transformation="exp", library_normalization="x_lib"):
        super().__init__()
        assert mean_transformation in ["exp", "softmax"]
        self.dispersion = dispersion
        self.mean_transformation = mean_transformation
        self.library_normalization = library_normalization

    @property
    def parameters(self):
        params = {
            "mean": "no_transformation",
        }
        if self.dispersion == "feature":
            params["r"] = "per_feature"
        else:
            raise NotImplementedError()
        return params

    def initial_transformation(self, x, x_mask=1.0):
        x = x
        aux_info = {}
        aux_info["x_library_size"] = calculate_library_size(x, x_mask)
        x = library_size_normalization(x, aux_info["x_library_size"], self.library_normalization)
        x = torch.log1p(x)
        return x, aux_info

    def dist(self, aux_info, parameters, lib_y):
        mean = parameters["mean"]
        r = 1.0 + parameters["r"]
        library_size = lib_y

        if self.mean_transformation == "exp":
            trans_mean = torch.exp(mean)
            trans_mean = library_size_correction(trans_mean, library_size, self.library_normalization, log_space=False)
        elif self.mean_transformation == "softmax":
            trans_mean = torch.softmax(mean, dim=-1)
            trans_mean = library_size.unsqueeze(-1) * trans_mean
        else:
            raise NotImplementedError()
        trans_r = torch.exp(r)
        return NegativeBinomial(mu=trans_mean, theta=trans_r, scale=None)


class LogNegativeBinomial(Distribution):
    def __init__(self, log_m, log_r, eps: float = 1e-8, validate_args=False) -> None:
        self.log_m = log_m
        self.log_r = log_r
        self._eps = eps
        super().__init__(validate_args=validate_args)

    @property
    def mean(self):
        return torch.exp(self.log_m)

    @property
    def theta(self):
        return torch.exp(self.log_r)

    @property
    def variance(self):
        return self.mean + (self.mean**2) / self.theta

    def sample(self, sample_shape):
        raise NotImplementedError()

    @staticmethod
    def negative_binomial_log_ver(k, m_log, r_log, eps=1e-8):
        # r :D
        r = torch.exp(r_log)

        # choice_part = log(binom(k, k+r-1))
        choice_part = torch.lgamma(k + r + eps) - torch.lgamma(k + 1 + eps) - torch.lgamma(r + eps)
        # log_pow_k = log(p ^ k) = log((m/(m+r)) ^ k)
        log_pow_k = -k * F.softplus(r_log - m_log + eps)
        # log_pow_r = log((1 - p) ^ r) = log((r/(m+r)) ^ r)
        log_pow_r = -r * F.softplus(m_log - r_log + eps)

        return choice_part + log_pow_k + log_pow_r

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        return self.negative_binomial_log_ver(value, self.log_m, self.log_r, eps=self._eps)


class LogNegativeBinomialNoiseModel(NoiseModel):
    def __init__(self, dispersion="feature", mean_transformation="none", library_normalization="x_lib"):
        super().__init__()
        self.dispersion = dispersion
        self.mean_transformation = mean_transformation
        self.library_normalization = library_normalization

    @property
    def parameters(self):
        params = {
            "mean": "no_transformation",
        }
        if self.dispersion == "feature":
            params["r"] = "per_feature"
        else:
            raise NotImplementedError()
        return params

    def initial_transformation(self, x, x_mask=1.0):
        x = x
        aux_info = {}
        aux_info["x_library_size"] = calculate_library_size(x, x_mask)
        x = library_size_normalization(x, aux_info["x_library_size"], self.library_normalization)
        x = torch.log1p(x)
        return x, aux_info

    def dist(self, aux_info, parameters, lib_y):
        mean = parameters["mean"]
        r = 1.0 + parameters["r"]
        library_size = lib_y

        if self.mean_transformation == "none":
            trans_mean = library_size_correction(mean, library_size, self.library_normalization, log_space=True)
        elif self.mean_transformation == "softmax":
            trans_mean = mean - torch.logsumexp(mean, dim=-1, keepdim=True)
            trans_mean = torch.log(library_size).unsqueeze(-1) + trans_mean
        else:
            raise NotImplementedError()
        trans_r = r
        return LogNegativeBinomial(log_m=trans_mean, log_r=trans_r)



import pickle

import numpy as np
import pandas as pd

from drvi.utils.metrics._aggregation import latent_matching_score, most_similar_averaging_score, most_similar_gap_score
from drvi.utils.metrics._pairwise import local_mutual_info_score, nn_alignment_score, spearman_correlataion_score

AVAILABLE_METRICS = {
    "ASC": spearman_correlataion_score,
    "SPN": nn_alignment_score,
    "SMI": local_mutual_info_score,
}


AVAILABLE_AGGREGATION_METHODS = {
    "LMS": latent_matching_score,
    "MSAS": most_similar_averaging_score,
    "MSGS": most_similar_gap_score,
}


class DiscreteDisentanglementBenchmark:
    version = "v1"

    def __init__(
        self,
        embed,
        discrete_target=None,
        one_hot_target=None,
        dim_titles=None,
        metrics=("SMI", "SPN", "ASC"),
        aggregation_methods=("LMS", "MSAS", "MSGS"),
    ):
        if discrete_target is None and one_hot_target is None:
            raise ValueError("Either discrete_target or one_hot_target must be provided.")
        if discrete_target is not None and one_hot_target is not None:
            raise ValueError("Only one of discrete_target or one_hot_target should be provided.")

        if discrete_target is not None:
            if isinstance(discrete_target, pd.Series):
                discrete_target = discrete_target.astype("category")
            elif isinstance(discrete_target, np.ndarray):
                discrete_target = pd.Series(discrete_target, dtype="category")
            else:
                raise ValueError("discrete_target must be a pandas Series or numpy array")
            one_hot_target = pd.DataFrame(
                np.eye(len(discrete_target.cat.categories))[discrete_target.cat.codes],
                columns=discrete_target.cat.categories,
            )

        if isinstance(one_hot_target, pd.DataFrame):
            pass
        elif isinstance(one_hot_target, np.ndarray):
            one_hot_target = pd.DataFrame(
                one_hot_target, columns=[f"process_{i}" for i in range(one_hot_target.shape[1])]
            )
        else:
            raise ValueError("one_hot_target must be a pandas DataFrame or numpy array")

        if dim_titles is None:
            dim_titles = [f"dim_{d}" for d in range(embed.shape[1])]

        self.embed = embed.copy()
        self.one_hot_target = one_hot_target.copy()
        self.dim_titles = dim_titles
        self.metrics = metrics
        self.aggregation_methods = aggregation_methods

        self.results = {}
        self.aggregated_results = {}

    @staticmethod
    def _compute_metrics(embed, one_hot_target, dim_titles=None, metrics=()):
        if dim_titles is None:
            dim_titles = [f"dim_{d}" for d in range(embed.shape[1])]

        results = {}
        for metric_name in metrics:
            result_df = pd.DataFrame(
                AVAILABLE_METRICS[metric_name](embed, gt_one_hot=one_hot_target.values),
                index=dim_titles,
                columns=one_hot_target.columns,
            )
            results[metric_name] = result_df

        return results

    @staticmethod
    def _aggregate_metrics(results, aggregation_methods=()):
        aggregated_results = {}
        for aggregation_method in aggregation_methods:
            for metric_name in results:
                aggregated_results[f"{aggregation_method}-{metric_name}"] = AVAILABLE_AGGREGATION_METHODS[
                    aggregation_method
                ](results[metric_name].values)
        return aggregated_results

    def is_complete(self):
        for metric in self.metrics:
            if metric not in self.results:
                return False
        for aggregation_method in self.aggregation_methods:
            for metric in self.metrics:
                if f"{aggregation_method}-{metric}" not in self.aggregated_results:
                    return False
        return True

    def evaluate(self):
        if not self.is_complete():
            remaining_metrics = [metric for metric in self.metrics if metric not in self.results]
            self.results = {
                **self.results,
                **self._compute_metrics(self.embed, self.one_hot_target, self.dim_titles, remaining_metrics),
            }
            # Aggregation is cheap. Do it always.
            self.aggregated_results = {
                **self.aggregated_results,
                **self._aggregate_metrics(self.results, self.aggregation_methods),
            }

    def get_results(self):
        return {
            f"{aggregation_method}-{metric}": self.aggregated_results[f"{aggregation_method}-{metric}"]
            for aggregation_method in self.aggregation_methods
            for metric in self.metrics
        }

    def get_results_details(self):
        return {f"{metric}": self.results[metric] for metric in self.metrics}

    def save(self, path):
        data = {
            "version": self.version,
            "results": self.results,
            "aggregated_results": self.aggregated_results,
            "metrics": self.metrics,
            "aggregation_methods": self.aggregation_methods,
            "dim_titles": self.dim_titles,
        }

        with open(path, "wb") as f:
            pickle.dump(data, f)

    @classmethod
    def load(cls, path, embed, discrete_target=None, one_hot_target=None, metrics=None, aggregation_methods=None):
        with open(path, "rb") as f:
            data = pickle.load(f)

        assert cls.version == data["version"]
        if metrics is None:
            metrics = data["metrics"]
        if aggregation_methods is None:
            aggregation_methods = data["aggregation_methods"]
        instance = cls(embed, discrete_target, one_hot_target, data["dim_titles"], metrics, aggregation_methods)
        instance.results = data["results"]
        instance.aggregated_results = data["aggregated_results"]
        return instance

    @classmethod
    def load_results(cls, path):
        with open(path, "rb") as f:
            data = pickle.load(f)
        return data["aggregated_results"]

    @classmethod
    def load_results_details(cls, path):
        with open(path, "rb") as f:
            data = pickle.load(f)
        return data["results"]



import numpy as np
import pandas as pd

from drvi.utils.metrics import DiscreteDisentanglementBenchmark


class TestDiscreteDisentanglementBenchmark:
    def make_test_data(self):
        categorical_features = pd.Series(pd.Categorical(np.repeat(["ct1", "ct2", "ct3"], 20)))
        categorical_features_01 = np.eye(len(categorical_features.cat.categories))[categorical_features.cat.codes]

        fit_continuous_latent = np.random.normal(categorical_features_01, 0.1)
        random_continuous_latent = np.random.normal(0, 1, fit_continuous_latent.shape)

        return categorical_features, categorical_features_01, fit_continuous_latent, random_continuous_latent

    def _general_test_pipeline(self, continuous_latent, categorical_features=None, categorical_features_01=None):
        METRICS = ["ASC", "SPN", "SMI"]
        AGGREGATION_METHODS = ["LMS", "MSAS", "MSGS"]

        benchmark = DiscreteDisentanglementBenchmark(
            continuous_latent,
            discrete_target=categorical_features,
            one_hot_target=categorical_features_01,
            metrics=METRICS,
            aggregation_methods=AGGREGATION_METHODS,
        )
        benchmark.evaluate()
        return benchmark

    def test_benchmarker_for_categorical_gt(self):
        categorical_features, categorical_features_01, fit_continuous_latent, random_continuous_latent = (
            self.make_test_data()
        )

        self._general_test_pipeline(
            continuous_latent=fit_continuous_latent,
            categorical_features=categorical_features,
        )

    def test_benchmarker_for_onehot_gt(self):
        categorical_features, categorical_features_01, fit_continuous_latent, random_continuous_latent = (
            self.make_test_data()
        )

        self._general_test_pipeline(
            continuous_latent=fit_continuous_latent, categorical_features_01=categorical_features_01
        )

    def test_benchmarker_make_sure_categorical_and_one_hot_have_same_result(self):
        categorical_features, categorical_features_01, fit_continuous_latent, random_continuous_latent = (
            self.make_test_data()
        )

        benchmark_1 = self._general_test_pipeline(
            continuous_latent=fit_continuous_latent, categorical_features_01=categorical_features_01
        )

        benchmark_2 = self._general_test_pipeline(
            continuous_latent=fit_continuous_latent, categorical_features_01=categorical_features_01
        )

        results_1 = benchmark_1.get_results()
        results_2 = benchmark_2.get_results()

        for metric in results_1:
            assert np.allclose(results_1[metric], results_2[metric]), f"Results for {metric} do not match"

    def test_benchmarker_make_sure_random_latent_have_worse_than_gt_results(self):
        categorical_features, categorical_features_01, fit_continuous_latent, random_continuous_latent = (
            self.make_test_data()
        )

        benchmark_gt = self._general_test_pipeline(
            continuous_latent=fit_continuous_latent,
            categorical_features=categorical_features,
        )

        benchmark_random = self._general_test_pipeline(
            continuous_latent=random_continuous_latent,
            categorical_features=categorical_features,
        )

        results_gt = benchmark_gt.get_results()
        results_random = benchmark_random.get_results()

        for metric in results_gt:
            assert np.all(results_random[metric] < results_gt[metric]), f"Results for {metric} do not match"



import anndata as ad
import numpy as np
import pandas as pd
from scipy import sparse

from drvi.model import DRVI


class TestDRVIModel:
    n = 1_000
    g = 200
    c = 2
    b = 5

    def make_test_adata(self, is_sparse=True):
        N, G, C, B = self.n, self.g, self.c, self.b

        ct_list = np.random.choice(range(C), N)[:, np.newaxis]
        batch_list = np.random.choice(range(B), N)[:, np.newaxis]
        batch_list_2 = np.random.choice(range(B), N)[:, np.newaxis]
        ct_array = (np.indices((N, C))[1] == ct_list) + 0.0
        g_exp_array = np.random.randint(0, 2, [G, C])
        exp_indicator = ct_array @ g_exp_array.T
        g_mean_list = np.exp(np.random.random(G) * 10 - 5)[:, np.newaxis]

        exp_matrix = np.random.poisson(exp_indicator * g_mean_list.T).astype(np.float32)

        adata = ad.AnnData(
            X=sparse.csr_matrix(exp_matrix),
            obs=pd.DataFrame(
                {
                    "cell_type": [f"ct_{ct}" for ct in ct_list[:, 0]],
                    "batch": [f"batch_{bid}" for bid in batch_list[:, 0]],
                    "batch_2": [f"batch_{bid}" for bid in batch_list_2[:, 0]],
                },
                index=[f"cell_{i}" for i in range(N)],
            ),
            var=pd.DataFrame(
                {
                    "gene_mean": g_mean_list[:, 0],
                    "gene_active_signature": np.apply_along_axis(
                        lambda x: "".join(x), axis=1, arr=g_exp_array.astype(str)
                    ),
                },
                index=[f"gene_{i}" for i in range(G)],
            ),
        )

        adata.obs["total_counts"] = adata.X.sum(axis=1)
        adata.layers["counts"] = adata.X.copy()
        adata.layers["lognorm"] = np.log1p(adata.X)

        if not is_sparse:
            adata.X = adata.X.A
            for l in ["counts", "lognorm"]:
                adata.layers[l] = adata.layers[l].A

        return adata

    def _general_integration_test(self, adata, layer="lognorm", data_kwargs=None, **kwargs):
        is_count_data = layer == "counts"
        setup_anndata_default_params = dict(  # noqa: C408
            categorical_covariate_keys=["batch"],
            layer=layer,
            is_count_data=is_count_data,
        )
        default_args = dict(  # noqa: C408
            n_latent=32,
            encoder_dims=[128],
            decoder_dims=[128],
            gene_likelihood="normal",
            categorical_covariates=["batch"],
            decoder_reuse_weights="everywhere",
        )
        DRVI.setup_anndata(adata, **{**setup_anndata_default_params, **(data_kwargs or {})})
        model = DRVI(adata, **{**default_args, **kwargs})
        print(model.module)
        model.train(accelerator="cpu", max_epochs=10)
        latent = model.get_latent_representation(adata)
        assert latent.shape[0] == adata.n_obs
        return {"model": model}

    def test_dimension_reduction_with_no_batch(self):
        adata = self.make_test_adata()
        self._general_integration_test(
            adata,
            categorical_covariates=[],
            data_kwargs=dict(categorical_covariate_keys=[]),  # noqa: C408
        )

    def test_simple_integration(self):
        adata = self.make_test_adata()
        self._general_integration_test(adata)

    def test_simple_integration_with_masking(self):
        adata = self.make_test_adata()
        self._general_integration_test(adata, fill_in_the_blanks_ratio=0.5)

    def test_simple_integration_latent_splitting(self):
        adata = self.make_test_adata()
        self._general_integration_test(adata, n_latent=32, n_split_latent=-1)
        self._general_integration_test(adata, n_latent=32, n_split_latent=8)
        self._general_integration_test(adata, n_latent=32, n_split_latent=8, split_method="power")
        self._general_integration_test(adata, n_latent=32, n_split_latent=8, split_method="split_map")
        self._general_integration_test(
            adata, n_latent=32, n_split_latent=8, split_method="split", split_aggregation="max"
        )

    def test_simple_integration_mean_activation(self):
        adata = self.make_test_adata()
        self._general_integration_test(adata, n_latent=32, n_split_latent=32, mean_activation="identity")
        self._general_integration_test(adata, n_latent=32, n_split_latent=32, mean_activation="relu")
        self._general_integration_test(adata, n_latent=32, n_split_latent=32, mean_activation="leaky_relu_0.4")
        self._general_integration_test(adata, n_latent=32, n_split_latent=32, mean_activation="elu_0.4")

    def test_decoder_reusing(self):
        adata = self.make_test_adata()
        for reuse_strategy in ["nowhere"]:
            self._general_integration_test(
                adata,
                n_latent=32,
                n_split_latent=8,
                split_method="split",
                split_aggregation="logsumexp",
                decoder_reuse_weights=reuse_strategy,
            )

    def test_integration_with_different_likelihoods(self):
        adata = self.make_test_adata()
        for gene_likelihood in ["nb", "normal", "pnb", "normal_sv"]:
            self._general_integration_test(
                adata,
                gene_likelihood=gene_likelihood,
                layer="counts" if gene_likelihood in ["nb", "pnb"] else "lognorm",
            )

    def test_integration_without_covariates(self):
        adata = self.make_test_adata()
        self._general_integration_test(
            adata,
            categorical_covariates=[],
            data_kwargs=dict(categorical_covariate_keys=[]),  # noqa: C408
        )

    def test_integration_with_different_covariate_modelings(self):
        adata = self.make_test_adata()
        for encode_covariates in [False, True]:
            for cms in [
                "one_hot",
                "emb_shared",
                "emb",
                "one_hot_linear",
                "emb_shared",
                "emb_linear",
            ]:
                self._general_integration_test(
                    adata, covariate_modeling_strategy=cms, encode_covariates=encode_covariates
                )

    def test_integration_with_different_var_activations(self):
        adata = self.make_test_adata()
        for var_activation in ["exp", "pow2"]:
            self._general_integration_test(adata, var_activation=var_activation)

    def test_integration_with_different_priors(self):
        adata = self.make_test_adata()
        for prior, prior_init_obs in [
            ("normal", None),
            ("gmm_5", None),
            ("gmm_5", adata.obs.index.to_series().sample(5)),
            ("vamp_5", adata.obs.index.to_series().sample(5)),
        ]:
            self._general_integration_test(adata, prior=prior, prior_init_obs=prior_init_obs)

    def test_multilevel_batch_integration(self):
        adata = self.make_test_adata()
        self._general_integration_test(
            adata,
            categorical_covariates=["batch", "batch_2"],
            data_kwargs=dict(categorical_covariate_keys=["batch", "batch_2"]),  # noqa: C408
        )
        self._general_integration_test(
            adata,
            categorical_covariates=["batch", "batch_2"],
            data_kwargs=dict(categorical_covariate_keys=["batch", "batch_2"]),  # noqa: C408
            prior="vamp_5",
            prior_init_obs=adata.obs.index.to_series().sample(5),
        )

    def _general_query_to_reference(self, adata_reference, adata_query, layer="lognorm", data_kwargs=None, **kwargs):
        is_count_data = layer == "counts"
        setup_anndata_default_params = dict(  # noqa: C408
            categorical_covariate_keys=["batch"],
            layer=layer,
            is_count_data=is_count_data,
        )
        default_args = dict(  # noqa: C408
            n_latent=32,
            encoder_dims=[128],
            decoder_dims=[128],
            gene_likelihood="normal",
            categorical_covariates=["batch"],
            decoder_reuse_weights="everywhere",
            encode_covariates=True,
        )
        DRVI.setup_anndata(adata_reference, **{**setup_anndata_default_params, **(data_kwargs or {})})
        model = DRVI(adata_reference, **{**default_args, **kwargs})
        model.train(accelerator="cpu", max_epochs=10)

        latent_reference = model.get_latent_representation(adata_reference)

        DRVI.prepare_query_anndata(adata_query, model)
        transfer_model = model.load_query_data(adata_query, model)
        transfer_model.train(accelerator="cpu", max_epochs=10, plan_kwargs={"lr": 0.1, "weight_decay": 0.0})
        latent_query = transfer_model.get_latent_representation(adata_query)
        transfer_model.train(accelerator="cpu", max_epochs=10, plan_kwargs={"lr": 0.1, "weight_decay": 0.0})
        latent_reference_after_train = transfer_model.get_latent_representation(adata_reference)
        latent_query_after_train = transfer_model.get_latent_representation(adata_query)

        assert np.sum((latent_reference - latent_reference_after_train) ** 2) < 1e-8
        assert np.sum((latent_query - latent_query_after_train) ** 2) > 1e-3
        assert latent_query_after_train.shape[0] == adata_query.n_obs

    def test_simple_query_to_reference_mapping(self):
        adata = self.make_test_adata()
        adata_reference = adata[adata.obs["batch"] != "batch_0"].copy()
        adata_query = adata[adata.obs["batch"] == "batch_0"].copy()
        self._general_query_to_reference(adata_reference, adata_query)

    def test_query_to_reference_mapping_with_different_cov_models(self):
        adata = self.make_test_adata()
        adata_reference = adata[adata.obs["batch"] != "batch_0"].copy()
        adata_query = adata[adata.obs["batch"] == "batch_0"].copy()
        for cms in [
            "one_hot",
            "emb_shared",
            "emb",
            "one_hot_linear",
            "emb_shared",
            "emb_linear",
        ]:
            self._general_query_to_reference(adata_reference, adata_query, covariate_modeling_strategy=cms)

    def test_query_to_reference_mapping_with_different_splits(self):
        adata = self.make_test_adata()
        adata_reference = adata[adata.obs["batch"] != "batch_0"].copy()
        adata_query = adata[adata.obs["batch"] == "batch_0"].copy()
        for n_split_latent in [1, 4, 32]:
            self._general_query_to_reference(adata_reference, adata_query, n_split_latent=n_split_latent)

    def test_query_to_reference_mapping_with_different_splits_and_different_cov_models(self):
        adata = self.make_test_adata()
        adata_reference = adata[adata.obs["batch"] != "batch_0"].copy()
        adata_query = adata[adata.obs["batch"] == "batch_0"].copy()
        for cms in ["one_hot", "emb_shared", "emb", "one_hot_linear", "emb_shared", "emb_linear"]:
            for n_split_latent in [1, 4, 32]:
                self._general_query_to_reference(
                    adata_reference, adata_query, n_split_latent=n_split_latent, covariate_modeling_strategy=cms
                )

    def test_continues_covariates(self):
        adata = self.make_test_adata()
        adata.obs["log_library_size_as_cont_cov"] = np.log1p(adata.layers["counts"].sum(-1))
        for encode_covariates in [False, True]:
            self._general_integration_test(
                adata,
                gene_likelihood="nb",
                layer="counts",
                encode_covariates=encode_covariates,
                data_kwargs=dict(continuous_covariate_keys=["log_library_size_as_cont_cov"]),  # noqa: C408
            )

    def test_ann_data_categorical_transfer_on_query_to_reference_mapping(self):
        # TODO: revise and complete this file
        adata = self.make_test_adata()
        adata_reference = adata[adata.obs["batch"] != "batch_0"].copy()
        adata_query = adata[adata.obs["batch"] == "batch_0"].copy()

        layer = "counts"
        is_count_data = layer == "counts"
        setup_anndata_default_params = dict(  # noqa: C408
            categorical_covariate_keys=["batch"],
            layer=layer,
            is_count_data=is_count_data,
        )
        default_args = dict(  # noqa: C408
            n_latent=32,
            encoder_dims=[128],
            decoder_dims=[128],
            gene_likelihood="normal",
            categorical_covariates=["batch"],
            decoder_reuse_weights="everywhere",
            encode_covariates=True,
        )
        DRVI.setup_anndata(adata_reference, **setup_anndata_default_params)
        model = DRVI(adata_reference, **default_args)
        model.train(accelerator="cpu", max_epochs=10)

        DRVI.prepare_query_anndata(adata_query, model)
        transfer_model = model.load_query_data(adata_query, model)
        transfer_model.train(accelerator="cpu", max_epochs=10, plan_kwargs={"lr": 0.1, "weight_decay": 0.0})

        for current_adata in (adata_query, adata_reference, adata):
            print("*****")
            current_adata = transfer_model._validate_anndata(current_adata)
            scdl = transfer_model._make_data_loader(
                adata=current_adata,
                indices=None,
                batch_size=128,
            )
            for tensors in scdl:
                print(tensors)
                inference_inputs = transfer_model.module._get_inference_input(tensors)
                print(inference_inputs)
                transfer_model.module.inference(**inference_inputs)
                break

    def test_reconstruction_of_a_latent_without_covariate(self):
        adata = self.make_test_adata()
        model = self._general_integration_test(
            adata,
            categorical_covariates=[],
            data_kwargs=dict(categorical_covariate_keys=[]),  # noqa: C408
        )["model"]

        latent = model.get_latent_representation(adata)
        reconstruction = model.decode_latent_samples(latent)
        assert reconstruction.shape == adata.X.shape

    def test_reconstruction_of_a_latent(self):
        adata = self.make_test_adata()
        model = self._general_integration_test(adata)["model"]
        cat_values = adata.obs[["batch"]].values

        latent = model.get_latent_representation(adata)
        reconstruction = model.decode_latent_samples(latent, cat_values=cat_values, map_cat_values=True)
        assert reconstruction.shape == adata.X.shape



import collections
import math
from collections.abc import Callable, Iterable, Sequence
from typing import Literal

import torch
from scvi.nn._utils import one_hot
from torch import nn
from torch.distributions import Normal

from drvi.nn_modules.embedding import MultiEmbedding
from drvi.nn_modules.freezable import FreezableBatchNorm1d, FreezableLayerNorm
from drvi.nn_modules.layer.factory import FCLayerFactory, LayerFactory
from drvi.nn_modules.layer.linear_layer import StackedLinearLayer
from drvi.nn_modules.noise_model import NoiseModel


def _identity(x):
    return x


class FCLayers(nn.Module):
    """A helper class to build fully-connected layers for a neural network.

    Parameters
    ----------
    layers_dim
        Number of nodes in layers including input and output dimensions
    n_cat_list
        A list containing, for each category of interest,
        the number of categories. Each category will be
        included using a one-hot encoding.
    dropout_rate
        Dropout rate to apply to each of the hidden layers
    split_size
        The size of split if input is a 3d tensor otherwise -1
        This parameter is required to handle batch normalization
    reuse_weights
        Whether to reuse weights when having multiple splits
    use_batch_norm
        Whether to have `BatchNorm` layers or not
    affine_batch_norm
        Whether to have affine transformation in `BatchNorm` layers
    use_layer_norm
        Whether to have `LayerNorm` layers or not
    use_activation
        Whether to have layer activation or not
    bias
        Whether to learn bias in linear layers or not
    inject_covariates
        Whether to inject covariates in each layer, or just the first (default).
    activation_fn
        Which activation function to use
    layer_factory
        A layer Factory instance to build projection layers based on
    layers_location
        An indicator to tell the class where in the architecture these layers reside.
    covariate_modeling_strategy
        The strategy model to consider covariates
    """

    def __init__(
        self,
        layers_dim: Sequence[int],
        n_cat_list: Iterable[int] = None,
        dropout_rate: float = 0.1,
        split_size: int = -1,
        reuse_weights: bool = True,
        use_batch_norm: bool = True,
        affine_batch_norm: bool = True,
        use_layer_norm: bool = False,
        use_activation: bool = True,
        bias: bool = True,
        inject_covariates: bool = True,
        activation_fn: nn.Module = nn.ELU,
        layer_factory: LayerFactory = None,
        layers_location: Literal["intermediate", "first", "last"] = "intermediate",
        covariate_modeling_strategy: Literal[
            "one_hot",
            "emb",
            "emb_shared",
            "one_hot_linear",
            "emb_linear",
            "emb_shared_linear",
        ] = "one_hot",
        covariate_embs_dim: Iterable[int] = (),
    ):
        super().__init__()
        self.inject_covariates = inject_covariates
        if covariate_modeling_strategy.endswith("_linear"):
            self.covariate_projection_modeling = "linear"
            self.covariate_vector_modeling = covariate_modeling_strategy[: -len("_linear")]
        else:
            self.covariate_projection_modeling = "cat"
            self.covariate_vector_modeling = covariate_modeling_strategy
        layer_factory = layer_factory or FCLayerFactory()

        self.n_cat_list = n_cat_list if n_cat_list is not None else []
        if self.covariate_vector_modeling == "one_hot":
            covariate_embs_dim = self.n_cat_list
        else:
            assert len(covariate_embs_dim) == len(self.n_cat_list)

        self.injectable_layers = []
        self.linear_batch_projections = []

        def is_intermediate(i):
            assert layers_location in ["intermediate", "first", "last"]
            if layers_location == "first" and i == 0:
                return False
            if layers_location == "last" and i == len(layers_dim) - 2:
                return False
            return True

        def inject_into_layer(layer_num):
            user_cond = layer_num == 0 or (layer_num > 0 and self.inject_covariates)
            return user_cond

        def get_projection_layer(n_in, n_out, i):
            output = []
            layer_needs_injection = False
            if not reuse_weights:
                assert split_size > 1
                if self.covariate_projection_modeling not in ["cat", "linear"]:
                    raise NotImplementedError()

            if len(self.n_cat_list) > 0 and inject_into_layer(i):
                layer_needs_injection = True
                cat_dim = sum(covariate_embs_dim)
                if self.covariate_vector_modeling == "emb":
                    batch_emb = MultiEmbedding(self.n_cat_list, covariate_embs_dim, init_method="normal", max_norm=1.0)
                    output.append(batch_emb)
                if self.covariate_projection_modeling == "cat":
                    n_in += cat_dim
                elif self.covariate_projection_modeling == "linear":
                    if reuse_weights:
                        linear_batch_projection = nn.Linear(cat_dim, n_out, bias=False)
                    else:
                        linear_batch_projection = StackedLinearLayer(split_size, cat_dim, n_out, bias=False)
                    output.append(linear_batch_projection)
                    self.linear_batch_projections.append(linear_batch_projection)
                else:
                    raise NotImplementedError()
            if reuse_weights:
                layer = layer_factory.get_normal_layer(
                    n_in,
                    n_out,
                    bias=bias,
                    intermediate_layer=is_intermediate(i),
                )
            else:
                layer = layer_factory.get_stacked_layer(
                    split_size,
                    n_in,
                    n_out,
                    bias=bias,
                    intermediate_layer=is_intermediate(i),
                )
            if layer_needs_injection:
                self.injectable_layers.append(layer)
            output.append(layer)
            return output

        def get_normalization_layers(n_out):
            output = []
            if split_size == -1:
                if use_batch_norm:
                    # non-default params come from defaults in original Tensorflow implementation
                    output.append(FreezableBatchNorm1d(n_out, momentum=0.01, eps=0.001, affine=affine_batch_norm))
                if use_layer_norm:
                    output.append(FreezableLayerNorm(n_out, elementwise_affine=False))
            else:
                if use_batch_norm:
                    # non-default params come from defaults in original Tensorflow implementation
                    output.append(
                        FreezableBatchNorm1d(n_out * split_size, momentum=0.01, eps=0.001, affine=affine_batch_norm)
                    )
                if use_layer_norm:
                    output.append(FreezableLayerNorm(n_out, elementwise_affine=False))
                    # The following logic is wrong
                    # output.append(FreezableLayerNorm([split_size, n_out], elementwise_affine=False))
            return output

        self.fc_layers = nn.Sequential(
            collections.OrderedDict(
                [
                    (
                        f"Layer {i}",
                        nn.ModuleList(
                            [
                                p
                                for p in [
                                    *get_projection_layer(n_in, n_out, i),
                                    *get_normalization_layers(n_out),
                                    activation_fn() if use_activation else None,
                                    nn.Dropout(p=dropout_rate) if dropout_rate > 0 else None,
                                ]
                                if p is not None
                            ]
                        ),
                    )
                    for i, (n_in, n_out) in enumerate(zip(layers_dim[:-1], layers_dim[1:], strict=True))
                ]
            )
        )

    def set_online_update_hooks(self, previous_n_cats_per_cov, n_cats_per_cov):
        """Set online update hooks."""
        if sum(previous_n_cats_per_cov) == sum(n_cats_per_cov):
            print("Nothing to make hook for!")
            return

        def make_hook_function(weight):
            w_size = weight.size()
            if weight.dim() == 2:
                # 2D tensors
                with torch.no_grad():
                    # Freeze gradients for normal nodes
                    if w_size[1] == sum(n_cats_per_cov):
                        transfer_mask = []
                    else:
                        transfer_mask = [
                            torch.zeros([w_size[0], w_size[1] - sum(n_cats_per_cov)], device=weight.device)
                        ]
                    # Iterate over the categories and Freeze old caterogies and make new ones trainable
                    for n_cat_new, n_cat_old in zip(n_cats_per_cov, previous_n_cats_per_cov, strict=False):
                        transfer_mask.append(torch.zeros([w_size[0], n_cat_old], device=weight.device))
                        if n_cat_new > n_cat_old:
                            transfer_mask.append(torch.ones([w_size[0], n_cat_new - n_cat_old], device=weight.device))
                    transfer_mask = torch.cat(transfer_mask, dim=1)
            elif weight.dim() == 3:
                # 3D tensors
                with torch.no_grad():
                    # Freeze gradients for normal nodes
                    if w_size[1] == sum(n_cats_per_cov):
                        transfer_mask = []
                    else:
                        transfer_mask = [
                            torch.zeros([w_size[0], w_size[1] - sum(n_cats_per_cov), w_size[2]], device=weight.device)
                        ]
                    # Iterate over the categories and Freeze old caterogies and make new ones trainable
                    for n_cat_new, n_cat_old in zip(n_cats_per_cov, previous_n_cats_per_cov, strict=False):
                        transfer_mask.append(torch.zeros([w_size[0], n_cat_old, w_size[2]], device=weight.device))
                        if n_cat_new > n_cat_old:
                            transfer_mask.append(
                                torch.ones([w_size[0], n_cat_new - n_cat_old, w_size[2]], device=weight.device)
                            )
                    transfer_mask = torch.cat(transfer_mask, dim=1)
            else:
                raise NotImplementedError()

            def _hook_fn_injectable(grad):
                return grad * transfer_mask

            return _hook_fn_injectable

        for layers in self.fc_layers:
            for layer in layers:
                if self.covariate_vector_modeling == "emb_shared":
                    # Nothing to do here :)
                    pass
                elif self.covariate_vector_modeling == "emb":
                    # Freeze everything but embs (new embs)
                    if isinstance(layer, MultiEmbedding):
                        assert tuple(layer.num_embeddings) == tuple(n_cats_per_cov)
                        print(f"Freezing old categories for {layer}")
                        layer.freeze_top_embs(previous_n_cats_per_cov)
                    # No need to handle others. For them required_grad is already set to False
                elif self.covariate_vector_modeling == "one_hot":
                    assert self.covariate_projection_modeling in ["cat", "linear"]
                    # Freeze everything but linears right after one_hot (new weights)
                    if (self.covariate_projection_modeling == "cat" and layer in self.injectable_layers) or (
                        self.covariate_projection_modeling == "linear" and layer in self.linear_batch_projections
                    ):
                        assert layer.weight.requires_grad
                        print(f"Registering backward hook parameter with shape {layer.weight.size()}")
                        layer.weight.register_hook(make_hook_function(layer.weight))
                        if layer.bias is not None:
                            assert not layer.bias.requires_grad
                else:
                    raise NotImplementedError()

    def forward(self, x: torch.Tensor, cat_full_tensor: torch.Tensor):
        """Forward computation on ``x``.

        Parameters
        ----------
        x
            tensor of values with shape ``(n_in,)``
        cat_full_tensor
            Tensor of category membership(s) for this sample

        Returns
        -------
        :class:`torch.Tensor`
            tensor of shape ``(n_out,)``
        """
        if self.covariate_vector_modeling == "one_hot":
            concat_list = []
            if cat_full_tensor is not None:
                cat_list = torch.split(cat_full_tensor, 1, dim=1)
            else:
                cat_list = ()

            if len(self.n_cat_list) > len(cat_list):
                raise ValueError("nb. categorical args provided doesn't match init. params.")
            for n_cat, cat in zip(self.n_cat_list, cat_list, strict=False):
                if n_cat and cat is None:
                    raise ValueError("cat not provided while n_cat != 0 in init. params.")
                concat_list += [one_hot(cat, n_cat)]
        elif self.covariate_vector_modeling == "emb_shared":
            concat_list = [cat_full_tensor]
        else:
            concat_list = []

        def dimension_transformation(t):
            if x.dim() == t.dim():
                return t
            if x.dim() == 3 and t.dim() == 2:
                return t.unsqueeze(dim=1).expand(-1, x.shape[1], -1)
            raise NotImplementedError()

        for layers in self.fc_layers:
            concat_list_layer = concat_list
            projected_batch_layer = None
            for layer in layers:
                if layer is not None:
                    if isinstance(layer, nn.BatchNorm1d):
                        if x.dim() == 3:
                            x = layer(x.reshape(x.shape[0], -1)).reshape(x.shape)
                        else:
                            x = layer(x)
                    elif isinstance(layer, MultiEmbedding):
                        assert self.covariate_vector_modeling in ["emb"]
                        assert len(concat_list) == 0
                        concat_list_layer = [layer(cat_full_tensor.int())]
                    elif layer in self.linear_batch_projections:
                        assert self.covariate_projection_modeling == "linear"
                        projected_batch_layer = layer(torch.cat(concat_list_layer, dim=-1))
                    else:
                        if layer in self.injectable_layers:
                            if self.covariate_projection_modeling == "cat":
                                current_cat_tensor = dimension_transformation(torch.cat(concat_list_layer, dim=-1))
                                x = torch.cat((x, current_cat_tensor), dim=-1)
                                x = layer(x)
                            elif self.covariate_projection_modeling in ["linear"]:
                                x = layer(x) + dimension_transformation(projected_batch_layer)
                            else:
                                raise NotImplementedError()
                        else:
                            x = layer(x)
        return x


# Encoder
class Encoder(nn.Module):
    """Encode data of ``n_input`` dimensions into a latent space of ``n_output`` dimensions.

    Uses a fully-connected neural network of ``n_hidden`` layers.

    Parameters
    ----------
    n_input
        The dimensionality of the input (data space)
    n_output
        The dimensionality of the output (latent space)
    layers_dim
        The number of nodes per hidden layer as a sequence
    n_cat_list
        A list containing the number of categories
        for each category of interest. Each category will be
        included using a one-hot encoding
    n_continuous_cov
        The number of continuous covariates
    inject_covariates
        Whether to inject covariates in each layer, or just the first (default).
    use_batch_norm
        Whether to use batch norm in layers
    affine_batch_norm
        Whether to use affine in batch norms
    use_layer_norm
        Whether to use layer norm in layers
    input_dropout_rate
        Dropout rate to apply to the input
    dropout_rate
        Dropout rate to apply to each of the hidden layers
    distribution
        Distribution of z
    var_eps
        Minimum value for the variance;
        used for numerical stability
    var_activation
        The activation function to ensure positivity of the variance. Defaults to "exp".
    mean_activation
        The activation function at the end of mean encoder. Defaults to "identity".
        Possible values are "identity", "relu", "leaky_relu", "leaky_relu_{slope}", "elu", "elu_{min_vaule}".
    layer_factory
        A layer Factory instance for building layers
    layers_location
        A hint on where the layer resides in the model
    covariate_modeling_strategy
        The strategy model takes to model covariates
    return_dist
        Return directly the distribution of z instead of its parameters.
    **kwargs
        Keyword args for :class:`~scvi.nn.FCLayers`
    """

    def __init__(
        self,
        n_input: int,
        n_output: int,
        layers_dim: Sequence[int] = (128,),
        n_cat_list: Iterable[int] = None,
        n_continuous_cov: int = 0,
        inject_covariates: bool = True,
        use_batch_norm: bool = True,
        affine_batch_norm: bool = True,
        use_layer_norm: bool = False,
        input_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        distribution: str = "normal",
        var_eps: float = 1e-4,
        var_activation: Callable | Literal["exp", "pow2", "2sig"] = "exp",
        mean_activation: Callable | str = "identity",
        layer_factory: LayerFactory = None,
        covariate_modeling_strategy: Literal[
            "one_hot",
            "emb",
            "emb_shared",
            "one_hot_linear",
            "emb_linear",
            "emb_shared_linear",
        ] = "one_hot",
        categorical_covariate_dims: Sequence[int] = (),
        return_dist: bool = False,
        **kwargs,
    ):
        super().__init__()

        self.distribution = distribution
        self.var_eps = var_eps
        self.input_dropout = nn.Dropout(p=input_dropout_rate)

        all_layers_dim = [n_input + n_continuous_cov] + list(layers_dim) + [n_output]
        if len(layers_dim) >= 1:
            self.encoder = FCLayers(
                layers_dim=all_layers_dim[:-1],
                n_cat_list=n_cat_list,
                dropout_rate=dropout_rate,
                use_batch_norm=use_batch_norm,
                affine_batch_norm=affine_batch_norm,
                use_layer_norm=use_layer_norm,
                inject_covariates=inject_covariates,
                layer_factory=layer_factory,
                layers_location="first",
                covariate_modeling_strategy=covariate_modeling_strategy,
                covariate_embs_dim=categorical_covariate_dims,
                **kwargs,
            )
        else:
            self.register_parameter("encoder", None)
            inject_covariates = True
        self.mean_encoder = FCLayers(
            layers_dim=all_layers_dim[-2:],
            n_cat_list=n_cat_list if inject_covariates else [],
            use_activation=False,
            use_batch_norm=False,
            use_layer_norm=False,
            bias=True,
            dropout_rate=0,
            layer_factory=layer_factory,
            layers_location="intermediate" if len(layers_dim) >= 1 else "first",
            covariate_modeling_strategy=covariate_modeling_strategy,
            covariate_embs_dim=categorical_covariate_dims if inject_covariates else [],
            **kwargs,
        )
        self.var_encoder = FCLayers(
            layers_dim=all_layers_dim[-2:],
            n_cat_list=n_cat_list if inject_covariates else [],
            use_activation=False,
            use_batch_norm=False,
            use_layer_norm=False,
            bias=True,
            dropout_rate=0,
            layer_factory=layer_factory,
            layers_location="intermediate" if len(layers_dim) >= 1 else "first",
            covariate_modeling_strategy=covariate_modeling_strategy,
            covariate_embs_dim=categorical_covariate_dims if inject_covariates else [],
            **kwargs,
        )
        self.return_dist = return_dist

        if distribution == "ln":
            self.z_transformation = nn.Softmax(dim=-1)
        else:
            self.z_transformation = _identity
        if var_activation == "exp":
            self.var_activation = torch.exp
        elif var_activation == "pow2":
            self.var_activation = lambda x: torch.pow(x, 2)
        elif var_activation == "2sig":
            self.var_activation = lambda x: 2 * torch.sigmoid(x)
        else:
            assert callable(var_activation)
            self.var_activation = var_activation

        if mean_activation == "identity":
            self.mean_activation = nn.Identity()
        elif mean_activation == "relu":
            self.mean_activation = nn.ReLU()
        elif mean_activation.startswith("leaky_relu"):
            if mean_activation == "leaky_relu":
                mean_activation = "leaky_relu_0.01"
            slope = float(mean_activation.split("leaky_relu_")[1])
            self.mean_activation = nn.LeakyReLU(negative_slope=slope)
        elif mean_activation.startswith("elu"):
            if mean_activation == "elu":
                mean_activation = "elu_1.0"
            alpha = float(mean_activation.split("elu_")[1])
            self.mean_activation = nn.ELU(alpha=alpha)
        elif mean_activation.startswith("celu"):
            if mean_activation == "celu":
                mean_activation = "celu_1.0"
            alpha = float(mean_activation.split("celu_")[1])
            self.mean_activation = nn.CELU(alpha=alpha)
        else:
            assert callable(mean_activation)
            self.mean_activation = mean_activation

    def forward(self, x: torch.Tensor, cat_full_tensor: torch.Tensor, cont_full_tensor: torch.Tensor = None):
        r"""The forward computation for a single sample.

         #. Encodes the data into latent space using the encoder network
         #. Generates a mean \\( q_m \\) and variance \\( q_v \\)
         #. Samples a new value from an i.i.d. multivariate normal \\( \\sim Ne(q_m, \\mathbf{I}q_v) \\)

        Parameters
        ----------
        x
            tensor with shape (n_input,)
        cat_full_tensor
            Tensor containing encoding of categorical variables of size n_batch x n_total_cat

        Returns
        -------
        3-tuple of :py:class:`torch.Tensor`
            tensors of shape ``(n_latent,)`` for mean and var, and sample

        """
        if cont_full_tensor is not None:
            x = torch.cat((x, cont_full_tensor), dim=-1)
        # Parameters for latent distribution
        q = self.encoder(self.input_dropout(x), cat_full_tensor) if self.encoder is not None else x
        q_m = self.mean_activation(self.mean_encoder(q, cat_full_tensor))
        q_v = self.var_activation(self.var_encoder(q, cat_full_tensor)) + self.var_eps
        dist = Normal(q_m, q_v.sqrt())
        latent = self.z_transformation(dist.rsample())
        if self.return_dist:
            return dist, latent
        return q_m, q_v, latent


# Decoder
class DecoderDRVI(nn.Module):
    """Decodes data from latent space of ``n_input`` dimensions into ``n_output`` dimensions.

    Uses a fully-connected neural network of ``n_hidden`` layers.

    Parameters
    ----------
    n_input
        The dimensionality of the input (latent space)
    n_output
        The dimensionality of the output (data space)
    layers_dim
        The number of nodes per hidden layer as a sequence
    n_cat_list
        A list containing the number of categories
        for each category of interest. Each category will be
        included using a one-hot encoding
    n_continuous_cov
        The number of continuous covariates
    n_split
        The number of splits for latent dim
    split_aggregation
        How to aggregate splits in the last layer of the decoder
    split_method
        How to make splits. Can be 'split' or 'power', or 'split_map'.
    reuse_weights
        Were to reuse the weights of the decoder layers when using splitting
        Possible values are 'everywhere', 'last', 'intermediate', 'nowhere'.
    dropout_rate
        Dropout rate to apply to each of the hidden layers
    inject_covariates
        Whether to inject covariates in each layer, or just the first (default).
    use_batch_norm
        Whether to use batch norm in layers
    affine_batch_norm
        Whether to use affine in batch norms
    use_layer_norm
        Whether to use layer norm in layers
    layer_factory
        A layer Factory instance for building layers
    covariate_modeling_strategy
        The strategy model takes to model covariates
    **kwargs
        Keyword args for :class:`~scvi.nn.FCLayers`.
    """

    def __init__(
        self,
        n_input: int,
        n_output: int,
        gene_likelihood_module: NoiseModel,
        n_cat_list: Iterable[int] = None,
        n_continuous_cov: int = 0,
        n_split: int = 1,
        split_aggregation: Literal["sum", "logsumexp", "max"] = "logsumexp",
        split_method: Literal["split", "power", "split_map"] = "split",
        reuse_weights: Literal["everywhere", "last", "intermediate", "nowhere"] = "everywhere",
        layers_dim: Sequence[int] = (128,),
        dropout_rate: float = 0.1,
        inject_covariates: bool = True,
        use_batch_norm: bool = False,
        affine_batch_norm: bool = True,
        use_layer_norm: bool = False,
        layer_factory: LayerFactory = None,
        covariate_modeling_strategy: Literal[
            "one_hot",
            "emb",
            "emb_shared",
            "one_hot_linear",
            "emb_linear",
            "emb_shared_linear",
        ] = "one_hot",
        categorical_covariate_dims: Sequence[int] = (),
        **kwargs,
    ):
        super().__init__()
        self.n_output = n_output
        self.gene_likelihood_module = gene_likelihood_module

        self.split_method = split_method
        self.n_split = n_split

        if n_split == -1 or n_split == 1:
            assert reuse_weights
            effective_dim = n_input
            self.n_split = n_split = -1
        elif self.split_method == "split":
            assert n_input % n_split == 0
            effective_dim = n_input // n_split
        elif self.split_method == "split_map":
            assert n_input % n_split == 0
            effective_dim = n_input
            self.split_transformation_weight = nn.Parameter(torch.randn(n_split, n_input // n_split, n_input))
            self.split_transformation_weight.data /= float(n_input // n_split) ** 0.5
        elif self.split_method == "power":
            effective_dim = n_input
            self.split_transformation = nn.Sequential(nn.Linear(n_input, n_input * n_split), nn.ReLU())
        else:
            raise NotImplementedError()

        self.effect_dim = effective_dim
        self.split_aggregation = split_aggregation

        assert reuse_weights in ["everywhere", "last", "intermediate", "nowhere"]
        intermediate_layers_reuse_weights = reuse_weights in ["everywhere", "intermediate"]
        last_layers_reuse_weights = reuse_weights in ["everywhere", "last"]

        all_layers_dim = [effective_dim + n_continuous_cov] + list(layers_dim) + [n_output]
        if len(layers_dim) >= 1:
            self.px_shared_decoder = FCLayers(
                layers_dim=all_layers_dim[:-1],
                split_size=n_split,
                reuse_weights=intermediate_layers_reuse_weights,
                n_cat_list=n_cat_list,
                dropout_rate=dropout_rate,
                inject_covariates=inject_covariates,
                use_batch_norm=use_batch_norm,
                affine_batch_norm=affine_batch_norm,
                use_layer_norm=use_layer_norm,
                layer_factory=layer_factory,
                layers_location="intermediate",
                covariate_modeling_strategy=covariate_modeling_strategy,
                covariate_embs_dim=categorical_covariate_dims,
                **kwargs,
            )
        else:
            self.register_parameter("px_shared_decoder", None)
            inject_covariates = True

        params_for_likelihood = self.gene_likelihood_module.parameters
        params_nets = {}
        for param_name, param_info in params_for_likelihood.items():
            if param_info.startswith("fixed="):
                params_nets[param_name] = torch.nn.Parameter(
                    torch.tensor(float(param_info.split("=")[1])), requires_grad=False
                )
            elif param_info == "no_transformation":
                params_nets[param_name] = FCLayers(
                    layers_dim=all_layers_dim[-2:],
                    split_size=n_split,
                    reuse_weights=last_layers_reuse_weights,
                    n_cat_list=n_cat_list if inject_covariates else [],
                    use_activation=False,
                    use_batch_norm=False,
                    use_layer_norm=False,
                    bias=True,
                    dropout_rate=0,
                    layer_factory=layer_factory,
                    layers_location="last",
                    covariate_modeling_strategy=covariate_modeling_strategy,
                    covariate_embs_dim=categorical_covariate_dims if inject_covariates else [],
                    **kwargs,
                )
            elif param_info == "per_feature":
                params_nets[param_name] = torch.nn.Parameter(torch.randn(n_output))
            else:
                raise NotImplementedError()
        self.params_nets = nn.ParameterDict(params_nets)

    def forward(
        self,
        z: torch.Tensor,
        cat_full_tensor: torch.Tensor,
        cont_full_tensor: torch.Tensor,
        library: torch.Tensor,
        gene_likelihood_additional_info: dict,
    ):
        """The forward computation for a single sample.

         #. Decodes the data from the latent space using the decoder network
         #. Returns distribution of expression
         #. If ``dispersion != 'gene-cell'`` then value for that param will be ``None``

        Parameters
        ----------
        z :
            tensor with shape ``(n_input,)``
        cat_full_tensor
            Tensor containing encoding of categorical variables of size n_batch x n_total_cat
        library
            library size
        gene_likelihood_additional_info
            additional info returned by gene likelihood module

        Returns
        -------
        distribution
        """
        batch_size = z.shape[0]
        if self.n_split > 1:
            if self.split_method == "power":
                z = self.split_transformation(z)
            z = torch.reshape(z, (batch_size, self.n_split, -1))
            if self.split_method == "split_map":
                z = torch.einsum("bsd,sdn->bsn", z, self.split_transformation_weight)

        if cont_full_tensor is not None:
            if self.n_split > 1:
                cont_full_tensor = cont_full_tensor.unsqueeze(1).expand(-1, self.n_split, -1)
            z = torch.cat((z, cont_full_tensor), dim=-1)

        last_tensor = self.px_shared_decoder(z, cat_full_tensor) if self.px_shared_decoder is not None else z
        original_params = {}
        params = {}
        for param_name, param_info in self.gene_likelihood_module.parameters.items():
            param_net = self.params_nets[param_name]
            if param_info.startswith("fixed="):
                original_params[param_name] = param_net
                params[param_name] = param_net.reshape(1, 1).expand(batch_size, self.n_output)
            elif param_info == "no_transformation":
                param_value = param_net(last_tensor, cat_full_tensor)
                original_params[param_name] = param_value
                if self.n_split > 1:
                    if self.split_aggregation == "sum":
                        # to get average
                        params[param_name] = param_value.sum(dim=-2) / self.n_split
                    elif self.split_aggregation == "logsumexp":
                        # to cancel the effect of n_splits
                        params[param_name] = torch.logsumexp(param_value, dim=-2) - math.log(self.n_split)
                    elif self.split_aggregation == "max":
                        params[param_name] = torch.amax(param_value, dim=-2)
                    else:
                        raise NotImplementedError()
                else:
                    params[param_name] = param_value
            elif param_info == "per_feature":
                original_params[param_name] = param_net
                params[param_name] = param_net.unsqueeze(0).expand(batch_size, -1)
            else:
                raise NotImplementedError()

        # Note this logic:
        px_dist = self.gene_likelihood_module.dist(
            aux_info=gene_likelihood_additional_info, parameters=params, lib_y=library
        )
        return px_dist, params, original_params



from torch import nn


def freezable(base_norm_class):
    class FreezableNormClass(base_norm_class):
        def __init__(self, *args, **kwargs):
            self._freeze = False
            super().__init__(*args, **kwargs)

        def freeze(self, freeze_status=True):
            self._freeze = freeze_status

        def forward(self, *args, **kwargs):
            training_status = self.training
            if self._freeze:
                self.train(False)
            result = super().forward(*args, **kwargs)
            self.train(training_status)
            return result

    return FreezableNormClass


@freezable
class FreezableBatchNorm1d(nn.BatchNorm1d):
    pass


@freezable
class FreezableLayerNorm(nn.LayerNorm):
    pass



import math

import torch
from torch import nn


class StackedLinearLayer(nn.Module):
    __constants__ = ["n_channels", "in_features", "out_features"]
    n_channels: int
    in_features: int
    out_features: int
    weight: torch.Tensor

    def __init__(
        self, n_channels: int, in_features: int, out_features: int, bias: bool = True, device=None, dtype=None
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.n_channels = n_channels
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.empty((n_channels, in_features, out_features), **factory_kwargs))
        if bias:
            self.bias = nn.Parameter(torch.empty(n_channels, out_features, **factory_kwargs))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        self._init_weight()
        if self.bias is not None:
            self._init_bias()

    def _init_weight(self) -> None:
        # Same as default nn.Linear (https://github.com/pytorch/pytorch/issues/57109)
        fan_in = self.in_features
        bound = 1 / math.sqrt(fan_in)
        nn.init.uniform_(self.weight, -bound, bound)

    def _init_bias(self) -> None:
        fan_in = self.in_features
        bound = 1 / math.sqrt(fan_in)
        nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        mm = torch.einsum("bci,cio->bco", input, self.weight)
        if self.bias is not None:
            mm = mm + self.bias  # They will broadcast well
        return mm

    def extra_repr(self) -> str:
        return (
            f"in_features={self.in_features}, out_features={self.out_features}, "
            f"n_channels={self.n_channels}, bias={self.bias is not None}"
        )
